(* Using alarms on the timer unit

    Timer base register: 4005,4000
    Setting the alarms:  4005,4010, etc.

Timer is chapter 4.6 from page 436 ff

Functions:
    ALARM00       ( +n -- )   - Print flags until time has elapsed
    ALARM-0       ( +n -- )   - Idem
    ONE           ( -- f )    - Give true flag when time has elapsed
    TWO           ( -- f )    - Idem
    START-ALARM0  ( -- )      - Start timer controlled flashing LED

*)

40054010 constant ALARM0    \ 00 04 08 0C   Four alarm cells 
40054020 constant ARMED     \ 01 02 04 08   Four armed flags
40054028 constant TIMERAWL  \ Low part of 64-bits timer
40054034 constant INTR      \ 01 02 04 08   Four interrupt flags


\ Primitive alarm functions
: ALARM00   ( +n -- )       \ Use alarm-0 with ARMED-flag
    timerawl @ +  alarm0 !
    begin
        armed @ 1 .r
    1 armed bit** 0= until
    cr ." Ready " armed @ . ;


: ALARM-0   ( +n -- )       \ Use alarm-0 with INTR-flag
    1 intr **bis
    timerawl @ +  alarm0 !
    begin
        intr @ 1 .r
    1 intr bit** until
    cr ." Ready " intr @ . ;


\ Defining word for alarm functions
: ALARM     ( +n +a -- f )  \ Define interval timer using the alarm function
    create  
        swap ,  3 umin          \ Alarm interval
        dup cells  alarm0 + ,   \ Used alarm
        bitmask ,               \ Bit masker
    does>
        dup 2 cells + @         \ Read bit mask
        armed bit** 0= dup if   \ Alarm not armed?
            drop  @+ timerawl @ +   \ Ok, calc next alarm time
            swap @ !  true  true    \ and set it, leave true
        then  nip ;             \ Remove data address

1000 1 alarm ONE    \ Define alarm-1 with 1000 cycles
8000 2 alarm TWO    \ Define alarm-2 with 8000 cycles

\ Test alarm
\
\   one . mamy      \ Test alarm-1
\   two . many      \ Test alarm-2



\ Timer alarm-0 interrupt example
code INT-ON     ( -- )      cpsie,  next, end-code
code INT-OFF    ( -- )      cpsid,  next, end-code  int-off

: IRQ!      ( a +n -- )         \ Set IRQ vector +n to a
    1F umin cells  40 +  ivecs +  vec! ;

dm 25 bitmask   constant GPIO-OUT   \ GPIO25 mask
1               constant ALARM0#    \ Alarm-0 mask
50000           constant PERIOD     \ Timer alarm period

routine FLASHES ( -- a )
    { hop day sun lr } push, \ Save used registers
    data>
        period ,    \ Alarm-0 cycle duration
        40054000 ,  \ Timer base; ALARM0=10, TIMERAWL=28, INTR=34
        gpio-out ,  \ Output bit mask
        D000001C ,  \ GPIO_OUT_XOR register
    code>
    w { hop day } ldm,  \ HOP=duration, DAY=timer-base
    sun  day 28 x) ldr, \ Read timer low
    sun hop add,        \ Add duration
    sun  day 10 x) str, \ Store next alarm-0 time
    sun alarm0# # movs, \ Reset alarm-0 interrupt flag
    sun  day 34 x) str,
    w { hop day } ldm,  \ HOP=GPIO-OUT, DAY=GPIO_XOR
    hop  day ) str,     \ Toggle GPIO-OUT
    { hop day sun pc } pop, \ Restore used registers
end-code

: START-ALARM0  ( -- )
    int-off                     \ Disable interrupts
    2000000 D0000020 !          \ GPIO_OE   set GPIO25 as output
    flashes  0 irq!             \ Install IRQ vector
    alarm0# 40054034 !          \ Clear alarm-0 interrupt flag
    alarm0# 40054038 !          \ Enable alarm-0 interrupt
    1 E000E100 !  int-on        \ NVIC_ISER enable timer interrupt 0 in NVIC
    timerawl @ period +  alarm0 ! ; \ Start alarm-0 of timer unit

start-alarm0

\ End
